<?php

declare(strict_types=1);

namespace Drupal\symfony_packages\Controller;

use Drupal\Core\Controller\ControllerBase;
use Drupal\Core\Render\RendererInterface;
use Drupal\symfony_packages\Form\SymfonyPackagesFilterForm;
use Drupal\symfony_packages\Service\SymfonyPackagesService;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\HttpFoundation\Request;

/**
 * Controller for the Symfony Packages report page.
 */
class SymfonyPackagesController extends ControllerBase {

  /**
   * The Symfony packages service.
   *
   * @var \Drupal\symfony_packages\Service\SymfonyPackagesService
   */
  protected SymfonyPackagesService $symfonyPackagesService;

  /**
   * The renderer service.
   *
   * @var \Drupal\Core\Render\RendererInterface
   */
  protected RendererInterface $renderer;

  /**
   * Constructs a SymfonyPackagesController object.
   *
   * @param \Drupal\symfony_packages\Service\SymfonyPackagesService $symfony_packages_service
   *   The Symfony packages service.
   * @param \Drupal\Core\Render\RendererInterface $renderer
   *   The renderer service.
   */
  public function __construct(
    SymfonyPackagesService $symfony_packages_service,
    RendererInterface $renderer,
  ) {
    $this->symfonyPackagesService = $symfony_packages_service;
    $this->renderer = $renderer;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    return new static(
      $container->get('symfony_packages.service'),
      $container->get('renderer')
    );
  }

  /**
   * Builds the Symfony packages report.
   *
   * @param \Symfony\Component\HttpFoundation\Request $request
   *   The current request.
   *
   * @return array
   *   A render array for the report page.
   */
  public function buildReport(Request $request): array {
    $build = [];

    // Clear cache if requested.
    if ($request->query->has('clear_cache')) {
      $this->symfonyPackagesService->clearCache();
      $this->messenger()->addMessage($this->t('Package cache has been cleared.'));
    }

    // Attach our CSS library.
    $build['#attached']['library'][] = 'symfony_packages/report';

    // Add the filter form.
    $filter_form = $this->formBuilder()->getForm(SymfonyPackagesFilterForm::class);
    $build['filter_form'] = $filter_form;

    // Get filter values.
    $filter = $request->query->get('filter', '');

    // Get package data, regardless of grouping.
    $packages = $this->symfonyPackagesService->getPackages($filter);

    // Main update report container.
    $build['symfony_packages'] = [
      '#type' => 'container',
      '#attributes' => ['class' => ['update']],
      'last_check' => [
        '#markup' => '',
        '#prefix' => '<div class="update__header">',
        '#suffix' => '</div>',
      ],
    ];

    // Add packages container.
    $build['symfony_packages']['packages'] = [
      '#type' => 'container',
      '#attributes' => ['class' => ['update__projects']],
    ];

    // Project container without any headings.
    $build['symfony_packages']['packages']['symfony'] = [
      '#type' => 'container',
      '#attributes' => ['class' => ['project-updates', 'project']],
    ];

    // Create the packages table.
    $rows = [];
    $header = [
      'name' => [
        'data' => $this->t('Package Name / Description'),
        'class' => ['project-update__name'],
      ],
      'version' => [
        'data' => $this->t('Current Version'),
        'class' => ['project-update__version'],
      ],
      'latest' => [
        'data' => $this->t('Latest Version'),
        'class' => ['project-update__latest-version'],
      ],
      'status' => [
        'data' => $this->t('Status'),
        'class' => ['project-update__status'],
      ],
    ];

    foreach ($packages as $package) {
      $status_class = $package['is_latest'] ? 'ok' : 'warning';
      $status_icon = $package['is_latest'] ? '✓' : '⚠';
      $status_text = $package['is_latest'] ? $this->t('Up to date') : $this->t('Update available');
      $description = empty($package['description']) ? $this->t('No description available') : $package['description'];

      // Create the package name cell with description underneath.
      $name_cell = [
        'data' => [
          'name' => [
            '#type' => 'html_tag',
            '#tag' => 'div',
            '#value' => '<a href="https://packagist.org/packages/' . $package['name'] . '" target="_blank">' . $package['name'] . '</a>',
            '#attributes' => ['class' => ['project-update__title']],
          ],
          'description' => [
            '#type' => 'html_tag',
            '#tag' => 'div',
            '#value' => $description,
            '#attributes' => ['class' => ['project-update__description']],
          ],
        ],
      ];

      // Create status icon and text.
      $status_cell = [
        'data' => [
          'icon' => [
            '#type' => 'html_tag',
            '#tag' => 'span',
            '#value' => $status_icon,
            '#attributes' => [
              'class' => [
                'project-update__status-icon',
                'project-update__status-icon--' . $status_class,
              ],
            ],
          ],
          'text' => [
            '#type' => 'html_tag',
            '#tag' => 'span',
            '#value' => $status_text,
            '#attributes' => ['class' => ['project-update__status-label']],
          ],
        ],
        'class' => ['project-update__status'],
      ];

      // Row classes for status.
      $row_class = ['project-update'];
      $row_class[] = $package['is_latest'] ? 'project-update--current' : 'project-update--not-current';

      $rows[] = [
        'data' => [
          'name' => $name_cell,
          'version' => [
            'data' => [
              '#markup' => '<a href="https://packagist.org/packages/' . $package['name'] . '#' . $package['version'] . '" target="_blank">' . $package['version'] . '</a>',
            ],
            'class' => ['project-update__current-version'],
          ],
          'latest' => [
            'data' => [
              '#markup' => '<a href="https://packagist.org/packages/' . $package['name'] . '#' . $package['latest_version'] . '" target="_blank">' . $package['latest_version'] . '</a>' .
              (!empty($package['prerelease']) ? '<div class="project-extra">' .
                  $this->t('Also available:') . ' <a href="https://packagist.org/packages/' . $package['name'] . '#' . $package['prerelease'] . '" target="_blank">' . $package['prerelease'] . '</a>' .
                '</div>' : ''),
            ],
            'class' => ['project-update__latest-version'],
          ],
          'status' => $status_cell,
        ],
        'class' => $row_class,
      ];
    }

    // Add the packages table.
    $build['symfony_packages']['packages']['symfony']['table'] = [
      '#type' => 'table',
      '#header' => $header,
      '#rows' => $rows,
      '#empty' => $this->t('No Symfony packages found.'),
      '#attributes' => [
        'class' => ['update-packages-table', 'project-updates__table'],
      ],
    ];

    return $build;
  }

}
