<?php

declare(strict_types=1);

namespace Drupal\symfony_packages\Service;

use Drupal\Core\Cache\CacheBackendInterface;
use Drupal\Core\File\FileSystemInterface;
use Drupal\Core\Logger\LoggerChannelFactoryInterface;
use Drupal\Core\Messenger\MessengerInterface;
use Drupal\Core\StringTranslation\StringTranslationTrait;
use GuzzleHttp\ClientInterface;
use Symfony\Component\Process\Exception\ProcessFailedException;
use Symfony\Component\Process\ExecutableFinder;
use Symfony\Component\Process\Process;

/**
 * Service for retrieving Symfony package information.
 */
class SymfonyPackagesService {

  use StringTranslationTrait;

  /**
   * The HTTP client.
   *
   * @var \GuzzleHttp\ClientInterface
   */
  protected ClientInterface $httpClient;

  /**
   * The file system service.
   *
   * @var \Drupal\Core\File\FileSystemInterface
   */
  protected FileSystemInterface $fileSystem;

  /**
   * The logger service.
   *
   * @var \Drupal\Core\Logger\LoggerChannelInterface
   */
  protected $logger;

  /**
   * The messenger service.
   *
   * @var \Drupal\Core\Messenger\MessengerInterface
   */
  protected MessengerInterface $messenger;

  /**
   * The cache service.
   *
   * @var \Drupal\Core\Cache\CacheBackendInterface
   */
  protected CacheBackendInterface $cache;

  /**
   * Cache of package descriptions.
   *
   * @var array
   */
  protected array $packageDescriptions = [];

  /**
   * List of Symfony packages required by Drupal core.
   *
   * @var array
   */
  protected array $corePackages = [];

  /**
   * Constructs a SymfonyPackagesService object.
   *
   * @param \GuzzleHttp\ClientInterface $http_client
   *   The HTTP client.
   * @param \Drupal\Core\File\FileSystemInterface $file_system
   *   The file system service.
   * @param \Drupal\Core\Logger\LoggerChannelFactoryInterface $logger_factory
   *   The logger factory service.
   * @param \Drupal\Core\Messenger\MessengerInterface $messenger
   *   The messenger service.
   * @param \Drupal\Core\Cache\CacheBackendInterface $cache
   *   The cache service.
   */
  public function __construct(
    ClientInterface $http_client,
    FileSystemInterface $file_system,
    LoggerChannelFactoryInterface $logger_factory,
    MessengerInterface $messenger,
    CacheBackendInterface $cache,
  ) {
    $this->httpClient = $http_client;
    $this->fileSystem = $file_system;
    $this->logger = $logger_factory->get('symfony_packages');
    $this->messenger = $messenger;
    $this->cache = $cache;
  }

  /**
   * Gets information about all installed Symfony packages.
   *
   * @param string $filter
   *   Optional filter to apply to package names or descriptions.
   *
   * @return array
   *   Array of package information.
   */
  public function getPackages(string $filter = ''): array {
    // Check cache first.
    $cid = 'symfony_packages:all';
    if ($cached = $this->cache->get($cid)) {
      $packages = $cached->data;

      // Apply filter if provided.
      if ($filter) {
        $packages = array_filter($packages, function ($package) use ($filter) {
          return str_contains(strtolower($package['name']), strtolower($filter)) ||
                 str_contains(strtolower($package['description']), strtolower($filter));
        });
      }

      return $packages;
    }

    // If not cached, build the data.
    $installed = $this->getInstalledPackages();
    $latest = $this->getLatestPackageVersions();
    $descriptions = $this->getPackageDescriptions($installed);
    $prereleases = $this->getPreReleaseVersions($installed);

    // Get Drupal core's required packages.
    $this->getCoreRequiredPackages();

    $packages = [];
    foreach ($installed as $name => $version) {
      // Skip non-Symfony packages.
      if (!str_starts_with($name, 'symfony/')) {
        continue;
      }

      $package = [
        'name' => $name,
        'version' => $version,
        'latest_version' => $latest[$name] ?? $version,
        'is_latest' => isset($latest[$name]) ? version_compare($version, $latest[$name], '>=') : TRUE,
        'description' => $descriptions[$name] ?? $this->t('No description available'),
        'prerelease' => $prereleases[$name] ?? NULL,
        'is_core' => in_array($name, $this->corePackages),
      ];

      $packages[$name] = $package;
    }

    // Sort by name.
    ksort($packages);

    // Cache the results for 1 hour.
    $this->cache->set($cid, $packages, time() + 3600);

    // Apply filter if provided.
    if ($filter) {
      $packages = array_filter($packages, function ($package) use ($filter) {
        return str_contains(strtolower($package['name']), strtolower($filter)) ||
               str_contains(strtolower($package['description']), strtolower($filter));
      });
    }

    return $packages;
  }

  /**
   * Gets a list of core and contrib packages separately.
   *
   * @param string $filter
   *   Optional filter to apply to package names or descriptions.
   *
   * @return array
   *   Array with 'core' and 'contrib' keys, each containing package arrays.
   */
  public function getGroupedPackages(string $filter = ''): array {
    $packages = $this->getPackages($filter);

    $grouped = [
      'core' => [],
      'contrib' => [],
    ];

    foreach ($packages as $name => $package) {
      if ($package['is_core']) {
        $grouped['core'][$name] = $package;
      }
      else {
        $grouped['contrib'][$name] = $package;
      }
    }

    return $grouped;
  }

  /**
   * Finds the composer.json directory by checking common locations.
   *
   * @return string|null
   *   The directory containing composer.json, or NULL if not found.
   */
  protected function findComposerJsonDirectory(): ?string {
    // Check parent directory of DRUPAL_ROOT (common in Composer-managed sites).
    $parentDir = dirname(DRUPAL_ROOT);
    if (file_exists($parentDir . '/composer.json')) {
      return $parentDir;
    }

    // Check DRUPAL_ROOT itself.
    if (file_exists(DRUPAL_ROOT . '/composer.json')) {
      return DRUPAL_ROOT;
    }

    // Check vendor directory for clues.
    $vendorDir = DRUPAL_ROOT . '/vendor';
    if (is_dir($vendorDir) && file_exists($vendorDir . '/composer/installed.json')) {
      // If vendor exists in web root, composer.json might be here too.
      return DRUPAL_ROOT;
    }

    return NULL;
  }

  /**
   * Gets information about installed packages using installed.json or Composer.
   *
   * @return array
   *   Array of installed packages with their versions.
   */
  protected function getInstalledPackages(): array {
    // Check cache first.
    $cid = 'symfony_packages:installed';
    if ($cached = $this->cache->get($cid)) {
      return $cached->data;
    }

    $packages = [];

    try {
      $composerDir = $this->findComposerJsonDirectory();
      if (!$composerDir) {
        $this->messenger->addError($this->t('Cannot find composer.json file. Make sure this is a Composer-managed Drupal installation.'));
        return $packages;
      }

      // Read installed.json directly instead of using Composer commands - much faster.
      $installedJsonPath = $composerDir . '/vendor/composer/installed.json';
      if (file_exists($installedJsonPath)) {
        $installedData = json_decode(file_get_contents($installedJsonPath), TRUE);
        if (isset($installedData['packages'])) {
          foreach ($installedData['packages'] as $package) {
            $packages[$package['name']] = $package['version'];

            // If package has a description, cache it directly to avoid additional lookups.
            if (isset($package['description'])) {
              $this->packageDescriptions[$package['name']] = $package['description'];
            }
          }

          // Cache the results for 1 hour.
          $this->cache->set($cid, $packages, time() + 3600);

          return $packages;
        }
      }

      // Fallback to composer command if needed.
      $finder = new ExecutableFinder();
      $composer = $finder->find('composer', 'composer');

      if (!$composer) {
        $this->messenger->addError($this->t('Cannot find Composer executable.'));
        return $packages;
      }

      $process = new Process([$composer, 'show', '--format=json', '--no-interaction']);
      $process->setWorkingDirectory($composerDir);
      // Increase timeout for large projects.
      $process->setTimeout(60);
      $process->run();

      if (!$process->isSuccessful()) {
        throw new ProcessFailedException($process);
      }

      $result = json_decode($process->getOutput(), TRUE);
      if (isset($result['installed'])) {
        foreach ($result['installed'] as $package) {
          $packages[$package['name']] = $package['version'];

          // If the package has a description, store it directly.
          if (isset($package['description'])) {
            $this->packageDescriptions[$package['name']] = $package['description'];
          }
        }

        // Cache the results for 1 hour.
        $this->cache->set($cid, $packages, time() + 3600);
      }
    }
    catch (\Exception $e) {
      $this->logger->error('Error getting installed packages: @message', ['@message' => $e->getMessage()]);
      $this->messenger->addError($this->t('Error getting installed packages: @message', ['@message' => $e->getMessage()]));
    }

    return $packages;
  }

  /**
   * Gets the latest available versions of Symfony packages.
   *
   * @return array
   *   Array of package names and their latest versions.
   */
  protected function getLatestPackageVersions(): array {
    // Check cache first.
    $cid = 'symfony_packages:latest';
    if ($cached = $this->cache->get($cid)) {
      return $cached->data;
    }

    $latest = [];

    try {
      $composerDir = $this->findComposerJsonDirectory();
      if (!$composerDir) {
        return $latest;
      }

      $finder = new ExecutableFinder();
      $composer = $finder->find('composer', 'composer');

      if (!$composer) {
        return $latest;
      }

      // Use a single process for all Symfony packages to reduce overhead.
      $process = new Process([$composer, 'show', '--latest', '--format=json', 'symfony/*', '--no-interaction']);
      $process->setWorkingDirectory($composerDir);
      // Increase timeout for large projects.
      $process->setTimeout(60);
      $process->run();

      if (!$process->isSuccessful()) {
        throw new ProcessFailedException($process);
      }

      $result = json_decode($process->getOutput(), TRUE);
      if (isset($result['installed'])) {
        foreach ($result['installed'] as $package) {
          if (isset($package['latest'])) {
            $latest[$package['name']] = $package['latest'];
          }
        }

        // Cache the results for 1 hour.
        $this->cache->set($cid, $latest, time() + 3600);
      }
    }
    catch (\Exception $e) {
      $this->logger->error('Error getting latest package versions: @message', ['@message' => $e->getMessage()]);
    }

    return $latest;
  }

  /**
   * Gets descriptions for Symfony packages.
   *
   * @param array $installed_packages
   *   The installed packages info to avoid duplicating work.
   *
   * @return array
   *   Array of package names and their descriptions.
   */
  protected function getPackageDescriptions(array $installed_packages): array {
    // Check if we already have descriptions from installed packages.
    if (!empty($this->packageDescriptions)) {
      return $this->packageDescriptions;
    }

    // Check cache.
    $cid = 'symfony_packages:descriptions';
    if ($cached = $this->cache->get($cid)) {
      $this->packageDescriptions = $cached->data;
      return $this->packageDescriptions;
    }

    try {
      $composerDir = $this->findComposerJsonDirectory();
      if (!$composerDir) {
        return $this->packageDescriptions;
      }

      $finder = new ExecutableFinder();
      $composer = $finder->find('composer', 'composer');

      if (!$composer) {
        return $this->packageDescriptions;
      }

      // Get only symfony packages.
      $symfonyPackages = [];
      foreach ($installed_packages as $name => $version) {
        if (str_starts_with($name, 'symfony/')) {
          $symfonyPackages[] = $name;
        }
      }

      if (empty($symfonyPackages)) {
        return $this->packageDescriptions;
      }

      // Use composer info to get all info at once instead of individual calls.
      $process = new Process(
        array_merge(
          [$composer, 'info', '--format=json', '--no-interaction'],
          $symfonyPackages
        )
      );
      $process->setWorkingDirectory($composerDir);
      // Increase timeout for large projects.
      $process->setTimeout(60);
      $process->run();

      if ($process->isSuccessful()) {
        $result = json_decode($process->getOutput(), TRUE);
        if (isset($result['packages'])) {
          foreach ($result['packages'] as $packageInfo) {
            if (isset($packageInfo['name']) && isset($packageInfo['description'])) {
              $this->packageDescriptions[$packageInfo['name']] = $packageInfo['description'];
            }
          }

          // Cache the results for 1 day (descriptions rarely change)
          $this->cache->set($cid, $this->packageDescriptions, time() + 86400);
        }
      }
    }
    catch (\Exception $e) {
      $this->logger->error('Error fetching package descriptions: @message', ['@message' => $e->getMessage()]);
    }

    return $this->packageDescriptions;
  }

  /**
   * Gets pre-release versions of Symfony packages.
   *
   * @param array $installed_packages
   *   The installed packages to check for pre-releases.
   *
   * @return array
   *   Array of package names and their pre-release versions.
   */
  protected function getPreReleaseVersions(array $installed_packages): array {
    // Check cache.
    $cid = 'symfony_packages:prereleases';
    if ($cached = $this->cache->get($cid)) {
      return $cached->data;
    }

    $prereleases = [];

    try {
      $composerDir = $this->findComposerJsonDirectory();
      if (!$composerDir) {
        return $prereleases;
      }

      $finder = new ExecutableFinder();
      $composer = $finder->find('composer', 'composer');

      if (!$composer) {
        return $prereleases;
      }

      // Process each package individually to find prereleases.
      foreach ($installed_packages as $packageName => $currentVersion) {
        if (!str_starts_with($packageName, 'symfony/')) {
          continue;
        }

        $process = new Process([$composer, 'show', $packageName, '-a', '--format=json', '--no-interaction']);
        $process->setWorkingDirectory($composerDir);
        $process->setTimeout(60);
        $process->run();

        if (!$process->isSuccessful()) {
          continue;
        }

        $result = json_decode($process->getOutput(), TRUE);
        if (!isset($result['versions']) || empty($result['versions'])) {
          continue;
        }

        // Sort versions in descending order.
        $versions = $result['versions'];
        usort($versions, function ($a, $b) {
          return version_compare($b, $a);
        });

        // Clean the current version by removing any 'v' prefix.
        $currentVersion = preg_replace('/^v/', '', $currentVersion);

        // Extract the major.minor version parts from current version.
        if (!preg_match('/^(\d+)\.(\d+)/', $currentVersion, $matches)) {
          continue;
        }

        $currentMajor = (int) $matches[1];
        $currentMinor = (int) $matches[2];

        // Find the newest prerelease for a newer major or minor version.
        foreach ($versions as $version) {
          // Only consider pre-release versions (alpha, beta, RC, DEV)
          if (!preg_match('/(alpha|beta|RC|DEV)/i', $version)) {
            continue;
          }

          // Extract major.minor from this version.
          if (!preg_match('/^v?(\d+)\.(\d+)/', $version, $vMatches)) {
            continue;
          }

          $major = (int) $vMatches[1];
          $minor = (int) $vMatches[2];

          // Check if this pre-release is for a newer major or minor version.
          if ($major > $currentMajor || ($major === $currentMajor && $minor > $currentMinor)) {
            $prereleases[$packageName] = $version;
            break;
          }
        }
      }

      // Cache the results for 6 hours.
      $this->cache->set($cid, $prereleases, time() + 21600);
    }
    catch (\Exception $e) {
      $this->logger->error('Error fetching pre-release versions: @message', ['@message' => $e->getMessage()]);
    }

    return $prereleases;
  }

  /**
   * Gets a list of Symfony packages required by Drupal core.
   */
  protected function getCoreRequiredPackages(): void {
    // Check cache first.
    $cid = 'symfony_packages:core_required';
    if ($cached = $this->cache->get($cid)) {
      $this->corePackages = $cached->data;
      return;
    }

    $this->corePackages = [];

    try {
      $composerDir = $this->findComposerJsonDirectory();
      if (!$composerDir) {
        return;
      }

      // Look for Drupal core's composer.json.
      $drupalCoreComposerJsonPath = '';

      // Check if we're in a Composer-managed Drupal site.
      if (file_exists($composerDir . '/web/core/composer.json')) {
        $drupalCoreComposerJsonPath = $composerDir . '/web/core/composer.json';
      }
      elseif (file_exists(DRUPAL_ROOT . '/core/composer.json')) {
        $drupalCoreComposerJsonPath = DRUPAL_ROOT . '/core/composer.json';
      }
      elseif (file_exists($composerDir . '/core/composer.json')) {
        $drupalCoreComposerJsonPath = $composerDir . '/core/composer.json';
      }

      if (empty($drupalCoreComposerJsonPath) || !file_exists($drupalCoreComposerJsonPath)) {
        // If we can't find Drupal core's composer.json, check the main composer.json for drupal/core.
        if (file_exists($composerDir . '/composer.json')) {
          $composerJson = json_decode(file_get_contents($composerDir . '/composer.json'), TRUE);
          if (isset($composerJson['require']['drupal/core'])) {
            // This is a drupal/core installation, look for the installed package.
            $vendorDir = $composerDir . '/vendor';
            if (file_exists($vendorDir . '/drupal/core/composer.json')) {
              $drupalCoreComposerJsonPath = $vendorDir . '/drupal/core/composer.json';
            }
          }
          elseif (isset($composerJson['require']['drupal/core-recommended'])) {
            // This is a drupal/core-recommended installation.
            $vendorDir = $composerDir . '/vendor';
            if (file_exists($vendorDir . '/drupal/core/composer.json')) {
              $drupalCoreComposerJsonPath = $vendorDir . '/drupal/core/composer.json';
            }
          }
        }
      }

      if (!empty($drupalCoreComposerJsonPath) && file_exists($drupalCoreComposerJsonPath)) {
        $coreComposerJson = json_decode(file_get_contents($drupalCoreComposerJsonPath), TRUE);

        // Extract Symfony dependencies.
        if (isset($coreComposerJson['require'])) {
          foreach ($coreComposerJson['require'] as $package => $constraint) {
            if (str_starts_with($package, 'symfony/')) {
              $this->corePackages[] = $package;
            }
          }
        }
      }
      else {
        // Fallback method: consider all common Symfony packages used by Drupal core.
        $commonCorePackages = [
          'symfony/console',
          'symfony/dependency-injection',
          'symfony/event-dispatcher',
          'symfony/http-foundation',
          'symfony/http-kernel',
          'symfony/routing',
          'symfony/serializer',
          'symfony/validator',
          'symfony/process',
          'symfony/yaml',
          'symfony/finder',
          'symfony/translation',
          'symfony/polyfill-iconv',
          'symfony/polyfill-php80',
          'symfony/polyfill-php81',
          'symfony/polyfill-ctype',
          'symfony/polyfill-intl-idn',
          'symfony/polyfill-mbstring',
        ];

        $this->corePackages = $commonCorePackages;
      }

      // Cache the results for 1 day.
      $this->cache->set($cid, $this->corePackages, time() + 86400);
    }
    catch (\Exception $e) {
      $this->logger->error('Error determining core Symfony packages: @message', ['@message' => $e->getMessage()]);
    }
  }

  /**
   * Clears the cached package information.
   */
  public function clearCache(): void {
    $this->cache->delete('symfony_packages:all');
    $this->cache->delete('symfony_packages:installed');
    $this->cache->delete('symfony_packages:latest');
    $this->cache->delete('symfony_packages:descriptions');
    $this->cache->delete('symfony_packages:prereleases');
    $this->cache->delete('symfony_packages:core_required');
    $this->packageDescriptions = [];
    $this->corePackages = [];
  }

}
